#!/usr/bin/env python3

import multiprocessing
import os
import queue
import shutil
import subprocess
import threading
import wave
from collections.abc import Iterable
from contextlib import contextmanager
from ctypes import string_at
from io import BytesIO

from rhvoice_wrapper import rhvoice_proxy

try:
    multiprocessing.Queue().qsize()
except NotImplementedError:
    OSX_FIX = True
else:
    OSX_FIX = False

_unset = object()
DEFAULT_CHUNK_SIZE = 1024 * 4
DEFAULT_FORMAT = 'wav'


class _WaveWrite(wave.Wave_write):
    def _ensure_header_written(self, _):
        pass

    def _patchheader(self):
        pass

    def write_header(self, init_length=0xFFFFFFF):   # Задаем 'бесконечную' длину файла
        self._write_header(init_length)


class _InOut(threading.Thread):
    def __init__(self, in_, out_, chunk_size):
        super().__init__()
        self._in = in_
        self._out = out_
        self._chunk_size = chunk_size
        self.start()

    def run(self):
        while True:
            try:
                chunk = self._in.read(self._chunk_size)
            except ValueError:
                break
            if not chunk:
                break
            self._out.put(chunk)


class _Pipe:
    # Pipe fasted for Process
    # Queue fasted for Thread (i don't know why)
    def __init__(self, is_multiprocessing=False):
        self._pipe = multiprocessing.Pipe(False) if is_multiprocessing else queue.Queue()
        if is_multiprocessing:
            self.get = self._pipe[0].recv
            self.put = self._pipe[1].send
        else:
            self.get = self._pipe.get
            self.put = self._pipe.put_nowait


class _StreamPipe:
    def __init__(self, is_multiprocessing=False):
        self._pipe = multiprocessing.Queue() if is_multiprocessing else queue.Queue()
        self.get = self._pipe.get
        self.put = self._pipe.put_nowait
        self.write = self.put
        if is_multiprocessing and OSX_FIX:
            self.qsize = self._osx_qsize
        else:
            self.qsize = self._pipe.qsize

    def _osx_qsize(self) -> int:
        return int(not self._pipe.empty())

    def clear(self):
        while self.qsize():
            try:
                self._pipe.get_nowait()
            except queue.Empty:
                pass

    def close(self):
        pass

    def flush(self):
        pass

    @staticmethod
    def tell():
        return 0


class _AudioWorker:
    SAMPLE_WIDTH = 2

    def __init__(self, cmd: dict, pipe: _StreamPipe):
        self._cmd = cmd
        self._stream = pipe
        self._wave = None
        self._starting = False

        self.get = self._stream.get
        self.qsize = self._stream.qsize

    def start_processing(self, format_, chunk_size, rate=24000):
        raise NotImplementedError

    def processing(self, samples, count):
        raise NotImplementedError

    def end_processing(self):
        raise NotImplementedError


class _AudioWorkerStream(_AudioWorker):
    POPEN_TIMEOUT = 10
    JOIN_TIMEOUT = 10

    def __init__(self, cmd: dict, pipe: _StreamPipe):
        super().__init__(cmd, pipe)
        self._in_out, self._popen = None, None

    def start_processing(self, format_, chunk_size, rate=24000):
        self._stream.clear()

        self._wave, self._in_out, self._popen = None, None, None

        if format_ != 'pcm':
            self._wave = _WaveWrite(self._select_target(format_, chunk_size))
            self._wave.setnchannels(1)
            self._wave.setsampwidth(self.SAMPLE_WIDTH)
            self._wave.setframerate(rate)
            self._wave.write_header()
        self._starting = True

    def processing(self, samples, count):
        data = string_at(samples, count * self.SAMPLE_WIDTH)
        if self._wave:
            self._wave.writeframesraw(data)
        else:
            self._stream.put(data)

    def end_processing(self):
        if not self._starting:
            # Генерации не было, надо отпустить клиента
            self._stream.put(b'')
            return False
        if self._wave:
            self._wave.close()
        if self._popen:
            self._popen.stdin.close()
            try:
                self._popen.wait(self.POPEN_TIMEOUT)
            except subprocess.TimeoutExpired:
                pass
        if self._in_out:
            self._in_out.join(timeout=self.JOIN_TIMEOUT)
        if self._popen:
            self._popen.stdout.close()
            self._popen.kill()
        self._stream.put(b'')
        self._starting = False
        return True

    def _create_popen(self, format_):
        self._popen = subprocess.Popen(
            self._cmd.get(format_),
            stdout=subprocess.PIPE,
            stdin=subprocess.PIPE
        )

    def _select_target(self, format_, chunk_size):
        if format_ in self._cmd:
            self._create_popen(format_)
            self._in_out = _InOut(self._popen.stdout, self._stream, chunk_size)
            return self._popen.stdin
        else:
            return self._stream


class _AudioWorkerBlocked(_AudioWorker):
    def __init__(self, cmd: dict, pipe: _StreamPipe):
        super().__init__(cmd, pipe)
        self._file, self._format = None, None

    def start_processing(self, format_, chunk_size, rate=24000):
        self._stream.clear()

        self._wave, self._format, self._file = None, format_, BytesIO()

        if self._format != 'pcm':
            self._wave = wave.Wave_write(self._file)
            self._wave.setnchannels(1)
            self._wave.setsampwidth(self.SAMPLE_WIDTH)
            self._wave.setframerate(rate)
        self._starting = True

    def processing(self, samples, count):
        data = string_at(samples, count * self.SAMPLE_WIDTH)
        if self._wave:
            self._wave.writeframesraw(data)
        else:
            self._file.write(data)

    def end_processing(self):
        if not self._starting:
            # Генерации не было, надо отпустить клиента
            self._stream.put(b'')
            return False
        if self._wave:
            self._wave.close()
        else:
            self._file.close()
        if self._format in self._cmd:
            try:
                self._stream.put(subprocess.check_output(self._cmd[self._format], input=self._file.getvalue()))
            except subprocess.CalledProcessError:
                pass
        else:
            self._stream.put(self._file.getvalue())
        self._file = None
        self._stream.put(b'')
        self._starting = False
        return True


class _BaseTTS:
    RELEASE_TIMEOUT = 3

    def __init__(self, is_multiprocessing: bool, free, cmd: dict, allow_formats: frozenset, **kwargs):
        _event = multiprocessing.Event if is_multiprocessing else threading.Event
        self._free = free
        self._allow_formats = allow_formats
        self._kwargs = kwargs.copy()
        self._is_stream = self._kwargs.pop('stream')
        self._lib_path = {} if 'lib_path' not in self._kwargs else {'lib_path': self._kwargs.pop('lib_path')}
        self._wait = _event()
        self._pipe = _Pipe(is_multiprocessing=is_multiprocessing)
        self._format = DEFAULT_FORMAT
        self._chunk_size = DEFAULT_CHUNK_SIZE
        self._engine = None
        _worker = _AudioWorkerStream if self._is_stream else _AudioWorkerBlocked
        self._worker = _worker(cmd=cmd, pipe=_StreamPipe(is_multiprocessing=is_multiprocessing))
        self._work = True
        self._client_here = _event()
        self._client_here.set()
        self._generator_work = _event()
        self._generator_work.set()
        self._still_processing = False

    def _engine_init(self):
        self._engine = rhvoice_proxy.Engine(**self._lib_path)
        self._engine.init(self._speech_callback, self._sr_callback, **self._kwargs)

    def _engine_destroy(self):
        self._engine.exterminate()
        self._engine = None

    def _speech_callback(self, samples, count, *_):
        self._worker.processing(samples, count)
        return not self._client_here.is_set() and self._work

    def _sr_callback(self, rate, *_):
        if not self._still_processing:
            self._still_processing = True
            self._worker.start_processing(self._format, self._chunk_size, rate)
            self._wait.set()
        return True

    def client_here(self):
        self._client_here.clear()

    def busy(self):
        return not (self._client_here.is_set() and self._generator_work.is_set())

    def _release_busy(self):
        # Wait while client reading data
        while not self._client_here.is_set():
            current_size = self._worker.qsize()
            self._client_here.wait(self.RELEASE_TIMEOUT)
            if current_size == self._worker.qsize():
                # Don't reading data? Client disconnected - set process as free
                break
        self._client_here.set()
        self._generator_work.set()
        self._free.set()

    def _client_request(self, text, voice, format_, chunk_size, sets):
        if format_ not in self._allow_formats:
            raise RuntimeError('Unsupported format: {}'.format(format_))
        sets = sets or {}
        if not isinstance(sets, dict):
            RuntimeError('Sets must be dict or None')
        if voice:
            sets['voice_profile'] = voice
        self._pipe.put((text, format_, chunk_size, sets))
        self._wait.wait(3600)
        self._wait.clear()

    @contextmanager
    def say(self, text, voice, format_, buff, sets):
        try:
            format_ = format_ or DEFAULT_FORMAT
            buff = buff if self._is_stream else None
            self._client_request(text, voice, format_, buff, sets)
            yield self._iter_me_splitting(buff) if format_ in ['pcm', 'wav'] and buff else self._iter_me()
        finally:
            self._client_here.set()

    def get(self, text, voice, format_, sets) -> bytes:
        format_ = format_ or DEFAULT_FORMAT
        try:
            self._client_request(text, voice, format_, DEFAULT_CHUNK_SIZE, sets)
            return b''.join(self._iter_me())
        finally:
            self._client_here.set()

    def to_file(self, filename, text, voice, format_, sets):
        try:
            with open(filename, 'wb') as fp:
                format_ = format_ or DEFAULT_FORMAT
                self._client_request(text, voice, format_, DEFAULT_CHUNK_SIZE, sets)
                for chunk in self._iter_me():
                    fp.write(chunk)
        finally:
            self._client_here.set()

    def set_params(self, **kwargs):
        self._pipe.put(kwargs)

    def _iter_me(self):
        while True:
            chunk = self._worker.get()
            if not chunk:
                break
            yield chunk

    def _iter_me_splitting(self, chunk_size):
        buffer = b''
        while True:
            chunk = self._worker.get()
            if not chunk:
                break
            buffer += chunk
            size = len(buffer)
            while size >= chunk_size:
                yield buffer[:chunk_size]
                buffer = buffer[chunk_size:]
                size -= chunk_size
        if buffer:
            yield buffer

    def _get_temporary_params(self, sets):
        try:
            return self._engine.params.copy_with(sets)
        except Exception as e:
            print('sets error: {}'.format(e))
        return None

    def _generate(self, text, format_, chunk_size, sets):
        self._generator_work.clear()
        self._format = format_
        self._chunk_size = chunk_size or DEFAULT_CHUNK_SIZE
        params = self._get_temporary_params(sets) if sets else None
        try:
            if isinstance(text, str):
                self._engine.generate(text=text, params=params)
            elif isinstance(text, Iterable):
                for chunk in text:
                    self._engine.generate(chunk, params=params)
        except RuntimeError:
            pass
        self._still_processing = False
        if not self._worker.end_processing():
            self._wait.set()
        self._release_busy()

    def run(self):
        self._engine_init()
        try:
            while self._work:
                data = self._pipe.get()
                if data is None:
                    break
                if isinstance(data, dict):
                    self._engine.set_params(**data)
                else:
                    self._generate(*data)
        finally:
            self._engine_destroy()

    def stop(self):
        if self._work:
            self._work = False
            self._pipe.put(None)


class ThreadTTS(_BaseTTS, threading.Thread):
    def __init__(self, *args, **kwargs):
        threading.Thread.__init__(self)
        _BaseTTS.__init__(self, False, *args, **kwargs)
        self.start()

    def join(self, timeout=None):
        self.stop()
        super().join()


class ProcessTTS(_BaseTTS, multiprocessing.Process):
    def __init__(self, *args, **kwargs):
        multiprocessing.Process.__init__(self)
        _BaseTTS.__init__(self, True, *args, **kwargs)
        self.start()

    def join(self, timeout=None):
        self.stop()
        super().join()


class MultiTTS:
    TIMEOUT = 30

    def __init__(self, count, processes, *args, **kwargs):
        if processes:
            worker = ProcessTTS
            self._free = multiprocessing.Event()
        else:
            worker = ThreadTTS
            self._free = threading.Event()
        self._workers = tuple([worker(self._free, *args, **kwargs) for _ in range(count)])
        self._lock = threading.Lock()
        self._work = True

    def to_file(self, filename: str, text: str, voice=None, format_=None, sets=None):
        """Generate and save audio in a file"""
        return self._caller().to_file(filename, text, voice, format_, sets)

    def say(self, text: str, voice=None, format_=None, buff=DEFAULT_CHUNK_SIZE, sets=None):
        """
        Starting audio generation and returned it chunk by chunk
        with tts.say(*args, **kwargs) as gen:
            print('chunks count: ', len([print('new chunk, len: ', len(chunk)) for chunk in gen]))
        """
        return self._caller().say(text, voice, format_, buff, sets)

    def get(self, text: str, voice=None, format_=None, sets=None) -> bytes:
        """Generate and returned audio as bytes"""
        return self._caller().get(text, voice, format_, sets)

    def _caller(self):
        with self._lock:
            while True:
                self._free.clear()
                for worker in self._workers:
                    if not worker.busy():
                        worker.client_here()
                        return worker
                if not self._free.wait(self.TIMEOUT):
                    raise RuntimeError('Still busy')

    def set_params(self, **kwargs):
        for worker in self._workers:
            worker.set_params(**kwargs)

    def join(self, *_):
        if not self._work:
            return
        self._work = False
        [x.stop() for x in self._workers]
        [x.join() for x in self._workers]


class TTS(MultiTTS):
    PARAMS = {
        'threads': 'THREADED',
        'force_process': 'PROCESSES_MODE',
        'lib_path': 'RHVOICELIBPATH',
        'data_path': 'RHVOICEDATAPATH',
        'resources': 'RHVOICERESOURCES',
        'lame_path': 'LAMEPATH',
        'opus_path': 'OPUSENCPATH',
        'flac_path': 'FLACPATH',
        'quiet': 'QUIET',
        'config_path': 'RHVOICECONFIGPATH',
        'stream': 'RHVOICESTREAM',
    }

    def __init__(self, threads=_unset, force_process=_unset,
                 lib_path=_unset, data_path=_unset, resources=_unset,
                 lame_path=_unset, opus_path=_unset, flac_path=_unset,
                 quiet=_unset, config_path=_unset, stream=_unset,
                 ):
        """
        :param int or bool or None threads: If equal to 1, created one thread object,
        if more running in multiprocessing mode and create a lot of processes. Default 1
        :param bool or None force_process: If True engines run in multiprocessing mode, if False in threads mode.
        Default False if threads == 1, else True.
        :param str or None lib_path: Path to RHVoice library.
        Default libRHVoice.so in Linux, libRHVoice.dylib in macOS and RHVoice.dll in Windows.
        :param str or None data_path: Path to folder, containing voices and languages folders.
        Default /usr/local/share/RHVoice.
        :param list or str or None resources: A list of paths to language and voice data.
        It should be used when it is not possible to collect all the data in one place. Default [].
        :param str or None lame_path: Path to lame, optional. File must be present for mp3 support. Default lame.
        :param str or None opus_path: Path to opusenc, optional. File must be present for opus support. Default opusenc.
        :param str or None flac_path: Path to flac, optional. File must be present for flac support. Default flac.
        :param bool or None quiet: If True don't info output. Default False.
        :param str or None config_path: Path to folder, contain RHVoice.conf in linux and RHVoice.ini in windows.
        Default /usr/local/etc/RHVoice.
        :param bool stream: Processing and sending chunks soon as possible,
        otherwise processing and sending only full data including length:
        say will return one big chunk, formats other than wav and pcm will be generated much slower. Default True.
        """
        envs = {}
        for key in self.PARAMS:
            if key in locals() and locals()[key] is not _unset:
                envs[key] = locals()[key]

        envs = self._get_environs(envs)
        quiet = self._prepare_bool(envs.pop('quiet', False))
        stream = self._prepare_bool(envs.pop('stream', True), True)
        self._threads = self._prepare_threads(envs.pop('threads', None))
        self._process = self._prepare_process(envs.pop('force_process', None), self._threads)
        self._cmd = self._get_cmd(
            quiet, stream,
            envs.pop('lame_path', None),
            envs.pop('opus_path', None),
            envs.pop('flac_path', None),
        )
        self._formats = frozenset(['pcm', 'wav'] + [key for key in self._cmd])

        self.__test_engine(envs.copy(), quiet)
        envs.update(stream=stream)
        super().__init__(self._threads, self._process, self._cmd, self._formats, **envs)

    def __test_engine(self, envs: dict, quiet: bool):
        lib_path = {} if 'lib_path' not in envs else {'lib_path': envs.pop('lib_path')}
        test = rhvoice_proxy.Engine(**lib_path)
        self._api = test.api
        self._version = test.version
        if self._version not in rhvoice_proxy.SUPPORT and not quiet:
            print(
                'Warning! Unsupported library version, use API {}. Supported: {}, library: {}.'.format(
                    self._api, rhvoice_proxy.SUPPORT, self._version
                )
            )
        test.init(play_speech_cb=lambda *_: True, set_sample_rate_cb=lambda *_: True, **envs)
        self._voices = test.voices
        self._params = test.params
        self._voice_profiles = test.voice_profiles
        test.exterminate()

    @property
    def formats(self) -> frozenset:
        return self._formats

    @property
    def thread_count(self) -> int:
        return self._threads

    @property
    def process(self) -> bool:
        return self._process

    @property
    def voices(self) -> tuple:
        return tuple([key for key in self._voices])

    @property
    def voice_profiles(self) -> tuple:
        return self._voice_profiles

    @property
    def api_version(self) -> str:
        return self._api

    @property
    def lib_version(self) -> str:
        return self._version

    @property
    def voices_info(self) -> dict:
        return self._voices

    @property
    def cmd(self) -> dict:
        return self._cmd

    def set_params(self, **kwargs) -> bool:
        result = False
        try:
            result = self._params.update_from_dict(kwargs)
        except RuntimeError as e:
            print('set_params error: {}'.format(e))
        if result:
            super().set_params(**kwargs)
        return result

    def get_params(self, param=None):
        if param is None:
            return self._params.to_dict()
        return self._params.get_param(param)

    def _get_environs(self, kwargs):
        result = {}
        for key, val in self.PARAMS.items():
            if key in kwargs:
                result[key] = kwargs[key]
            elif val in os.environ:
                result[key] = os.environ[val]
        return result

    @staticmethod
    def _prepare_bool(val, def_: bool = False):
        if isinstance(val, str):
            val = val.lower()
            if val in ['true', 'yes', 'enable']:
                return True
            elif val in ['false', 'no', 'disable']:
                return False
        elif isinstance(val, bool):
            return val
        return def_

    def _prepare_process(self, force_process, threads):
        return self._prepare_bool(force_process, threads > 1)

    @staticmethod
    def _prepare_threads(threads):
        if threads is None:
            return 1
        if isinstance(threads, bool):
            if threads:
                return multiprocessing.cpu_count()
            else:
                return 1
        try:
            threads = int(threads)
        except ValueError:
            threads = 1
        else:
            threads = threads if threads > 0 else 1
        return threads

    @staticmethod
    def _get_cmd(quiet, stream, lame, opus, flac):
        base_cmd = {
            'mp3': [[lame or 'lame', '-t', '-hv', '--silent', '-', '-'], 'lame'],
            'opus': [[opus or 'opusenc', '--ignorelength', '--quiet', '--discard-comments', '-', '-'], 'opus-tools'],
            'flac': [[flac or 'flac', '--ignore-chunk-sizes', '--totally-silent', '--best', '--stdout', '-'], 'flac'],
        }
        cmd = {}

        for key, val in base_cmd.items():
            if shutil.which(val[0][0]):
                cmd[key] = val[0]
                if not stream:
                    cmd[key].pop(1)
            elif not quiet:
                print('Disable {} support - {} not found. Use apt install {}'.format(key, val[0][0], val[1]))
        return cmd



#  part of rhvoice_bindings.py

from ctypes import c_char_p, byref

from rhvoice_wrapper.rhvoice_bindings import (
    RHVoice_callbacks, RHVoice_callback_types, RHVoice_init_params, RHVoice_synth_params, RHVoice_message_type,
    RHVoice_punctuation_mode, RHVoice_capitals_mode, RHVoice_synth_flag, load_tts_library, get_rhvoice_version
)

try:
    # noinspection PyUnresolvedReferences
    import rhvoice_wrapper_bin
    _LIB_PATH = rhvoice_wrapper_bin.lib_path
    _DATA_PATH = rhvoice_wrapper_bin.data_path
except ImportError:
    _LIB_PATH = None
    _DATA_PATH = None

SUPPORT = ('0.7.2', '1.0.0', '1.2.0')


class _SynthesisCheck:
    # magic?
    MIN_BASE = -2
    MAX_BASE = 2.5
    # from 0 to X, int
    MAX_PUNCTUATION = 3
    MAX_CAPITALS = 4
    MAX_FLAGS = 1
    #
    DEFAULT_VOICE = 'Anna'

    @staticmethod
    def __pass_interval(value, min_, max_):
        if value > max_:
            raise TypeError('{} > {}'.format(value, max_))
        if value < min_:
            raise TypeError('{} < {}'.format(value, min_))
        return value

    @classmethod
    def _pass_voice(cls, voices: str or list) -> bytes:
        if not isinstance(voices, list):
            voices = [voices]
        voices = '+'.join([x.capitalize() for x in voices[:2] if x])
        return (voices or cls.DEFAULT_VOICE).encode()

    @classmethod
    def _pass_base(cls, value: int or float) -> int or float:
        """MIN_BASE <= val <= MAX_BASE"""
        return cls.__pass_interval(value, cls.MIN_BASE, cls.MAX_BASE)

    @classmethod
    def __pass_enum(cls, value: int, max_: int) -> int:
        """0 <= val <= max_"""
        if not isinstance(value, int):
            raise TypeError('Must be int, get {}'.format(type(value)))
        return cls.__pass_interval(value, 0, max_)

    @classmethod
    def _pass_punctuation_mode(cls, value: int) -> int:
        return cls.__pass_enum(value, cls.MAX_PUNCTUATION)

    @classmethod
    def _pass_capitals_mode(cls, value: int) -> int:
        return cls.__pass_enum(value, cls.MAX_CAPITALS)

    @classmethod
    def _pass_flags(cls, value: int) -> int:
        return cls.__pass_enum(value, cls.MAX_FLAGS)

    @classmethod
    def _pass_punctuation_list(cls, punctuation_list: str or None) -> bytes or None:
        return punctuation_list.encode() if punctuation_list else None


class SynthesisParams(_SynthesisCheck):
    CHECKS = {
        # key: (default value, validation+clear function)
        'voice_profile': (_SynthesisCheck.DEFAULT_VOICE.encode(), _SynthesisCheck._pass_voice),
        'absolute_rate': (0, _SynthesisCheck._pass_base),
        'absolute_pitch': (0, _SynthesisCheck._pass_base),
        'absolute_volume': (0, _SynthesisCheck._pass_base),
        'relative_rate': (1, _SynthesisCheck._pass_base),
        'relative_pitch': (1, _SynthesisCheck._pass_base),
        'relative_volume': (1, _SynthesisCheck._pass_base),
        'punctuation_mode': (RHVoice_punctuation_mode.default, _SynthesisCheck._pass_punctuation_mode),
        'punctuation_list': (None, _SynthesisCheck._pass_punctuation_list),
        'capitals_mode': (RHVoice_capitals_mode.default, _SynthesisCheck._pass_capitals_mode),
        'flags': (int(not RHVoice_synth_flag.dont_clip_rate), _SynthesisCheck._pass_flags),
    }

    def __init__(self, api: tuple, params: dict or None = None):
        self.api = api
        self.synth_params = RHVoice_synth_params(self.api)()
        if params:
            self.update_from_dict(params)
        else:
            self._set_default()

    def update_from_dict(self, params: dict) -> bool:
        is_change = False
        for key, value in params.items():
            try:
                old_value = getattr(self.synth_params, key)
            except AttributeError:
                continue
            if key not in self.CHECKS:
                continue
            try:
                new_value = self.CHECKS[key][1](value)
            except Exception as e:
                raise RuntimeError('Wrong value from {}: {}'.format(key, e))
            if old_value == new_value:
                continue
            setattr(self.synth_params, key, new_value)
            is_change = True
        return is_change

    def get_param(self, key):
        try:
            return self._get_param(key)
        except AttributeError:
            return None

    def to_dict(self) -> dict:
        return {key: self._get_param(key) for key in self._synth_params_keys()}

    def copy_with(self, params: dict):
        result = SynthesisParams(self.api, params=self.to_dict())
        if params:
            result.update_from_dict(params)
        return result

    def _set_default(self):
        for key in self._synth_params_keys():
            setattr(self.synth_params, key, self.CHECKS[key][0])

    def _get_param(self, key):
        result = getattr(self.synth_params, key)
        return result.decode() if isinstance(result, bytes) else result

    def _synth_params_keys(self):
        # noinspection PyProtectedMember
        for key, _ in self.synth_params._fields_:
            yield key


class Engine:
    DEFAULT_DATA_PATH = '/usr/local/share/RHVoice'
    DEFAULT_CONFIG_PATH = '/usr/local/etc/RHVoice'
    GENDERS = {0: 'unknown', 1: 'male', 2: 'female'}

    def __init__(self, lib_path=_LIB_PATH, api=None):
        self._lib, self._api = load_tts_library(lib_path, api)
        self.params = SynthesisParams(self._api)
        self._engine, self.__save_me = None, None

    @property
    def version(self):
        return get_rhvoice_version(self._lib)

    @property
    def api(self):
        return '.'.join(str(k) for k in self._api)

    def init(self, play_speech_cb, set_sample_rate_cb, resources=None, data_path=_DATA_PATH, config_path=None):
        """initialize speech engine - load language data and set callbacks."""
        resources = resources or []
        if isinstance(resources, str):
            resources = [resources]
        resource_paths = [k.encode() for k in resources]

        callbacks = RHVoice_callbacks(self._api)(
            play_speech=RHVoice_callback_types.play_speech(play_speech_cb),
            set_sample_rate=RHVoice_callback_types.set_sample_rate(set_sample_rate_cb)
        )
        # noinspection PyTypeChecker
        params = RHVoice_init_params(self._api)(
            resource_paths=(c_char_p * (len(resource_paths) + 1))(*(resource_paths + [None])),
            data_path=(data_path or self.DEFAULT_DATA_PATH).encode(),
            config_path=(config_path or self.DEFAULT_CONFIG_PATH).encode(),
            callbacks=callbacks
        )
        self._engine = self._lib.RHVoice_new_tts_engine(byref(params))
        if not self._engine:
            raise RuntimeError('RHVoice: engine initialization error')
        # link for params must be present in memory while engine works
        self.__save_me = params

    @property
    def voices(self) -> dict:
        """
        Returns nested dictionary with voice information. First
        level key is voice name in lowercase, second level keys
        are voice properties.
        """
        voices = dict()
        voices_raw = self._lib.RHVoice_get_voices(self._engine)
        for number in range(self._lib.RHVoice_get_number_of_voices(self._engine)):
            vi = voices_raw[number]
            try:
                name = vi.name.decode()
            except (UnicodeDecodeError, AttributeError) as e:
                print('Wrong voice name, ignore #{}: {}'.format(number, e))
                continue
            key = name.lower()
            voices[key] = {
                'no': number,
                'name': name,
                'lang': vi.language.decode(errors='replace') if vi.language else 'NaN',
                'gender': self.GENDERS.get(vi.gender, 'NaN'),
                'country': 'NaN' if self._api < (1, 0, 0) or not vi.country else vi.country.decode(errors='replace')
            }
        return voices

    @property
    def voice_profiles(self) -> tuple:
        result = list()
        profiles_raw = self._lib.RHVoice_get_voice_profiles(self._engine)
        for number in range(self._lib.RHVoice_get_number_of_voice_profiles(self._engine)):
            try:
                result.append(profiles_raw[number].decode())
            except UnicodeDecodeError:
                pass
        return tuple(result)

    def generate(self, text, params: SynthesisParams = None):
        text = text.encode()
        synth_params = (params or self.params).synth_params
        message = self._lib.RHVoice_new_message(
            self._engine,
            text,
            len(text),
            RHVoice_message_type.text,
            byref(synth_params),
            None
        )
        if not message:
            raise RuntimeError('RHVoice: message building error')
        self._lib.RHVoice_speak(message)
        self._lib.RHVoice_delete_message(message)  # free the memory (check when message is stored)

    def set_params(self, **kw):
        self.params.update_from_dict(kw)

    def exterminate(self):
        if self._engine:
            self._lib.RHVoice_delete_tts_engine(self._engine)
            self._engine = None
            self.__save_me = None



import subprocess

from setuptools import setup


def get_version() -> str:
    version_file = 'version'

    def version_to_file(ver):
        with open(version_file, mode='w') as fd:
            fd.write(ver)

    def version_from_file():
        with open(version_file) as fd:
            return fd.read().splitlines()[0]

    def version_from_git():
        cmd = ['git', 'describe', '--abbrev=0', '--tags']
        try:
            return subprocess.check_output(cmd).decode().splitlines()[0]
        except Exception as e:
            print('ERROR! Execute {}: {}'.format(cmd, e))
            return None
    version = version_from_git()
    if not version:
        version = version_from_file()
        print('WARNING! Get version from a file: {}'.format(version))
    else:
        version_to_file(version)
    return version


def get_long_description():
    with open('README.md') as fh:
        return fh.read()


setup(
    name='rhvoice-wrapper',
    version=get_version(),
    packages=['rhvoice_wrapper'],
    url='https://github.com/Aculeasis/rhvoice-proxy',
    license='GPLv3+',
    author='Aculeasis',
    author_email='amilpalimov2@ya.ru',
    description='High-level interface for RHVoice library',
    long_description=get_long_description(),
    long_description_content_type='text/markdown',
    python_requires='>=3.4',
    extras_require={
        'rhvoice': ['rhvoice-wrapper-bin'],
    },
    classifiers=[
        'Intended Audience :: Developers',
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
        'Operating System :: OS Independent',
        'Topic :: Multimedia :: Sound/Audio :: Speech',
        'Topic :: Software Development :: Libraries',
    ],
)



# Copyright (C) 2010-2013  Olga Yakovleva <yakovleva.o.v@gmail.com>
# Copyright (C) 2015  anatoly techtonik <techtonik@gmail.com>
# Copyright (C) 2018 Aculeasis

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Base on:
# https://github.com/Olga-Yakovleva/RHVoice/blob/master/src/nvda-synthDriver/__init__.py
# https://github.com/techtonik/RHVoice/blob/master/src/nvda-synthDriver/RHVoice.py

__author__ = "Olga Yakovleva <yakovleva.o.v@gmail.com>"

import os
import platform
from ctypes import CDLL, CFUNCTYPE, POINTER, Structure, c_char_p, c_double, c_int, c_uint, c_short, c_void_p


class RHVoice_tts_engine_struct(Structure):
    pass


RHVoice_tts_engine = POINTER(RHVoice_tts_engine_struct)


class RHVoice_message_struct(Structure):
    pass


RHVoice_message = POINTER(RHVoice_message_struct)


class RHVoice_callback_types:
    set_sample_rate = CFUNCTYPE(c_int, c_int, c_void_p)
    play_speech = CFUNCTYPE(c_int, POINTER(c_short), c_uint, c_void_p)
    process_mark = CFUNCTYPE(c_int, c_char_p, c_void_p)
    word_starts = CFUNCTYPE(c_int, c_uint, c_uint, c_void_p)
    word_ends = CFUNCTYPE(c_int, c_uint, c_uint, c_void_p)
    sentence_starts = CFUNCTYPE(c_int, c_uint, c_uint, c_void_p)
    sentence_ends = CFUNCTYPE(c_int, c_uint, c_uint, c_void_p)
    play_audio = CFUNCTYPE(c_int, c_char_p, c_void_p)
    done = CFUNCTYPE(None, c_void_p)


class RHVoice_callbacks_072(Structure):
    _fields_ = [("set_sample_rate", RHVoice_callback_types.set_sample_rate),
                ("play_speech", RHVoice_callback_types.play_speech),
                ("process_mark", RHVoice_callback_types.process_mark),
                ("word_starts", RHVoice_callback_types.word_starts),
                ("word_ends", RHVoice_callback_types.word_ends),
                ("sentence_starts", RHVoice_callback_types.sentence_starts),
                ("sentence_ends", RHVoice_callback_types.sentence_ends),
                ("play_audio", RHVoice_callback_types.play_audio)]


class RHVoice_callbacks_100(Structure):
    # noinspection PyProtectedMember
    _fields_ = RHVoice_callbacks_072._fields_.copy()
    _fields_.append(("done", RHVoice_callback_types.done))


def RHVoice_callbacks(api: tuple):
    if api < (1, 0, 0):
        return RHVoice_callbacks_072
    return RHVoice_callbacks_100


class RHVoice_init_params_072(Structure):
    _fields_ = [("data_path", c_char_p),
                ("config_path", c_char_p),
                ("resource_paths", POINTER(c_char_p)),
                ("callbacks", RHVoice_callbacks_072),
                ("options", c_uint)]


class RHVoice_init_params_100(Structure):
    _fields_ = [("data_path", c_char_p),
                ("config_path", c_char_p),
                ("resource_paths", POINTER(c_char_p)),
                ("callbacks", RHVoice_callbacks_100),
                ("options", c_uint)]


def RHVoice_init_params(api: tuple):
    if api < (1, 0, 0):
        return RHVoice_init_params_072
    return RHVoice_init_params_100


class RHVoice_voice_info_072(Structure):
    _fields_ = [("language", c_char_p),
                ("name", c_char_p),
                ("gender", c_int)]


class RHVoice_voice_info_100(Structure):
    # noinspection PyProtectedMember
    _fields_ = RHVoice_voice_info_072._fields_.copy()
    _fields_.append(("country", c_char_p))


def RHVoice_voice_info(api: tuple):
    if api < (1, 0, 0):
        return RHVoice_voice_info_072
    return RHVoice_voice_info_100


class RHVoice_synth_params_100(Structure):
    _fields_ = [("voice_profile", c_char_p),
                ("absolute_rate", c_double),
                ("absolute_pitch", c_double),
                ("absolute_volume", c_double),
                ("relative_rate", c_double),
                ("relative_pitch", c_double),
                ("relative_volume", c_double),
                ("punctuation_mode", c_int),
                ("punctuation_list", c_char_p),
                ("capitals_mode", c_int)]


class RHVoice_synth_params_120(Structure):
    # noinspection PyProtectedMember
    _fields_ = RHVoice_synth_params_100._fields_.copy()
    _fields_.append(("flags", c_int))


def RHVoice_synth_params(api: tuple):
    if api < (1, 2, 0):
        return RHVoice_synth_params_100
    return RHVoice_synth_params_120


class RHVoice_message_type:
    text = 0
    ssml = 1
    characters = 2


class RHVoice_voice_gender:
    unknown = 0
    male = 1
    female = 2


class RHVoice_punctuation_mode:
    default = 0
    none = 1
    all = 2
    some = 3


class RHVoice_capitals_mode:
    default = 0
    off = 1
    word = 2
    pitch = 3
    sound = 4


class RHVoice_synth_flag:
    dont_clip_rate = 1


def get_rhvoice_version(lib):
    return lib.RHVoice_get_version().decode('utf-8')


def _get_compatible_api(lib) -> tuple:
    # noinspection PyBroadException
    try:
        version = tuple(int(x) for x in get_rhvoice_version(lib).split('.'))
    except Exception:
        version = (1, 0, 0)

    if version < (1, 0, 0):
        return 0, 7, 2
    if version < (1, 2, 0):
        return 1, 0, 0
    return 1, 2, 0


def _lib_selector(lib_path):
    if lib_path is None:
        if os.name == 'nt':
            lib_path = 'RHVoice.dll'
        else:
            lib_path = 'libRHVoice.{}'.format('dylib' if platform.system().lower() == 'darwin' else 'so')
    return lib_path if os.name == 'nt' else lib_path.encode()


def load_tts_library(lib_path=None, api=None):
    lib = CDLL(_lib_selector(lib_path))
    lib.RHVoice_get_version.restype = c_char_p
    api = api or _get_compatible_api(lib)
    lib.RHVoice_new_tts_engine.argtypes = (POINTER(RHVoice_init_params(api)),)
    lib.RHVoice_new_tts_engine.restype = RHVoice_tts_engine
    lib.RHVoice_delete_tts_engine.argtypes = (RHVoice_tts_engine,)
    lib.RHVoice_delete_tts_engine.restype = None
    lib.RHVoice_get_number_of_voices.argtypes = (RHVoice_tts_engine,)
    lib.RHVoice_get_number_of_voices.restype = c_uint
    lib.RHVoice_get_voices.argtypes = (RHVoice_tts_engine,)
    lib.RHVoice_get_voices.restype = POINTER(RHVoice_voice_info(api))
    lib.RHVoice_get_number_of_voice_profiles.argtypes = (RHVoice_tts_engine,)
    lib.RHVoice_get_number_of_voice_profiles.restype = c_uint
    lib.RHVoice_get_voice_profiles.argtypes = (RHVoice_tts_engine,)
    lib.RHVoice_get_voice_profiles.restype = POINTER(c_char_p)
    lib.RHVoice_are_languages_compatible.argtypes = (RHVoice_tts_engine, c_char_p, c_char_p)
    lib.RHVoice_are_languages_compatible.restype = c_int
    lib.RHVoice_new_message.argtypes = (
        RHVoice_tts_engine, c_char_p, c_uint, c_int, POINTER(RHVoice_synth_params(api)), c_void_p)
    lib.RHVoice_new_message.restype = RHVoice_message
    lib.RHVoice_delete_message.arg_types = (RHVoice_message,)
    lib.RHVoice_delete_message.restype = None
    lib.RHVoice_speak.argtypes = (RHVoice_message,)
    lib.RHVoice_speak.restype = c_int
    return lib, api
